/******************************************************************************
*
* project name:    TI-Mahjongg Solitaire
* file name:       timah.c
* initial date:    05/11/2000
* authors:         thomas.nussbaumer@gmx.net
*                  marcos.lopez@wol.es
*
* description:     main program
*
* compile with: tigcc -O2 timah.c
*
*
* NOTE:  No animals or trees were harmed for the stones of this Mahjongg game
*
*
* $Id: timah.c,v 1.1 2000/11/14 19:14:01 Thomas Nussbaumer Exp $
*
******************************************************************************/

#define SAVE_SCREEN
#define OPTIMIZE_ROM_CALLS
#define NO_EXIT_SUPPORT

#define USE_TI89
#define USE_TI92PLUS
#define USE_V200

#include <tigcclib.h>


#define MAHJONGG_VERSION "v1.10"

#include "board.c"
#include "cursor.c"
#include "clock.c"

/*---------------------------------------------------------------------------*/
/* some globals                                                              */
/*---------------------------------------------------------------------------*/
SCR_RECT fullscreen = {0,0,239,127}; // helper variable for clipping


/*---------------------------------------------------------------------------*/
/* some defines for drawing functions                                        */
/*---------------------------------------------------------------------------*/
#define COLOR_WHITE      0
#define COLOR_LIGHTGRAY  1
#define COLOR_DARKGRAY   2
#define COLOR_BLACK      3
#define LIGHT_PLANE      0
#define DARK_PLANE       1
#define RECT_EMPTY       0
#define RECT_FILLED      1
#define A_CENTERED       0x40
#define A_SHADOWED       0x80

int cursor_x = 0;
int cursor_y = 0;


/*===========================================================================*/
/* utility funcction to draw a string to both planes                         */
/* NOTE: there exists to new styles which are handled as flags               */
/*       A_CENTERED ... the given font will be used and the string will be   */
/*                      drawn centered on the screen                         */
/*       A_SHADOWED ... if this flag is set in the attributes a lightgray    */
/*                      shadow will be drawn with +1 pixel offset in all     */
/*                      two directions                                       */
/*===========================================================================*/
void DrawString(int x,int y,char* s,int font,int attr) {
    int do_shadow = attr & A_SHADOWED;

    x += offsetx;
    y += offsety;

    FontSetSys(font);

    if (attr & A_CENTERED) {
        x = (LCD_WIDTH-DrawStrWidth(s,font))/2;
        attr &= ~A_CENTERED;
    }

    attr &= ~A_SHADOWED;

    SetPlane(DARK_PLANE);
    DrawStr(x,y,s,attr);
    SetPlane(LIGHT_PLANE);
    DrawStr(x,y,s,attr);

    if (do_shadow) DrawStr(x+1,y+1,s,A_NORMAL);
}


/*===========================================================================*/
/* draw a line with given color from (x0,y0) to (x1,y1)                      */
/*===========================================================================*/
void DrawColorLine(int x0,int y0,int x1, int y1, int color) {
    SetPlane(LIGHT_PLANE);
    if (color == COLOR_WHITE || color == COLOR_DARKGRAY) {
        DrawLine(x0, y0, x1, y1, A_REVERSE);
    }
    else {
        DrawLine(x0, y0, x1, y1, A_NORMAL);
    }
    SetPlane(DARK_PLANE);
    if (color == COLOR_WHITE || color == COLOR_LIGHTGRAY) {
        DrawLine(x0, y0, x1, y1, A_REVERSE);
    }
    else {
        DrawLine(x0, y0, x1, y1, A_NORMAL);
    }
}


/*===========================================================================*/
/* draws a rectangle with given color                                        */
/* if fill == 1) the rectangle is filled                                     */
/*===========================================================================*/
void DrawColorRect(int x0,int y0,int x1, int y1, int color,int fill) {
    WIN_RECT wrect;
    SCR_RECT rect;
    void     (*df)(void*,void*,int);
    void*    parea;

    if (fill) {
        rect.xy.x0 = x0;
        rect.xy.y0 = y0;
        rect.xy.x1 = x1;
        rect.xy.y1 = y1;
        parea      = &rect;
        df         = (void(*)(void*,void*,int))ScrRectFill;
    }
    else {
        wrect.x0 = x0;
        wrect.y0 = y0;
        wrect.x1 = x1;
        wrect.y1 = y1;
        parea    = &wrect;
        df       = (void(*)(void*,void*,int))DrawClipRect;
    }

    SetPlane(LIGHT_PLANE);
    if (color == COLOR_WHITE || color == COLOR_DARKGRAY) {
        (*df)(parea,&fullscreen, A_REVERSE);
    }
    else {
        (*df)(parea,&fullscreen, A_NORMAL);
    }
    SetPlane(DARK_PLANE);
    if (color == COLOR_WHITE || color == COLOR_LIGHTGRAY) {
        (*df)(parea,&fullscreen, A_REVERSE);
    }
    else {
        (*df)(parea,&fullscreen, A_NORMAL);
    }
}


/*===========================================================================*/
/* draws a popup window                                                      */
/*===========================================================================*/
void DrawPopup(int x0,int y0,int x1, int y1) {
    DrawColorRect(x0,y0,x1,y1,COLOR_BLACK,RECT_EMPTY);
    x0++,y0++,x1--,y1--;
    DrawColorRect(x0,y0,x1,y1,COLOR_DARKGRAY,RECT_EMPTY);
    x0++,y0++,x1--,y1--;
    DrawColorRect(x0,y0,x1,y1,COLOR_LIGHTGRAY,RECT_EMPTY);
    x0++,y0++,x1--,y1--;
    DrawColorRect(x0,y0,x1,y1,COLOR_WHITE,RECT_FILLED);
}


/*===========================================================================*/
/* outputs the time to the statusline                                        */
/*===========================================================================*/
void PrintTime(void) {
    char buffer[16];
    sprintf(buffer,"%02d:%02d:%02d", hours, minutes, seconds);
    DrawString(130,95,buffer,F_4x6,A_REPLACE);
}


/*===========================================================================*/
/* get input key pressed by user                                             */
/*                                                                           */
/* replacement for ngetchx() and/or GKeyIn(), because:                       */
/*                                                                           */
/* (1) ngetchx() won't check the APD timer value                             */
/* (2) GKeyIn() destroys the display due to setting the status display       */
/*                                                                           */
/* (handles also time printing in its inner loop)                            */
/*===========================================================================*/
int GetUserInput() {
    void *kbq = kbd_queue();
    int key;

    //-------------------------------------------
    // restart the APD timer now
    //-------------------------------------------
    OSTimerRestart(APD_TIMER);

    while (1) {
        //---------------------------------------
        // wait until we get a key
        // if the APD timer expires -> turn
        // calculator off
        //---------------------------------------
        while (OSdequeue(&key, kbq)) {
            if (OSTimerExpired(APD_TIMER)) {
                off();
                OSTimerRestart(APD_TIMER);
            }
            if (outputtime) {
                outputtime = 0;
                PrintTime();
            }
        }

        //---------------------------------------
        // handle calculator-off key combinations
        // if the APD timer expires -> turn
        // calculator off
        //---------------------------------------
        if (key == 4363 || (TI89 && key == 16651) || (!TI89 && key == 8459)) {
            off();
            OSTimerRestart(APD_TIMER);
        }
        else {
            return key & 0xf7ff;  // mask out 0x800 (repeat flag)
        }
    }
}


#include "logo.c"

/*===========================================================================*/
/* treatment of cursor position changement                                   */
/*===========================================================================*/
void CursorUpdate(void) {
    int toplevel;
    int pos;

    if (cursor_x >= BOARD_SIZE_X) cursor_x = 0;
    if (cursor_x < 0)             cursor_x = BOARD_SIZE_X-1;
    if (cursor_y >= BOARD_SIZE_Y) cursor_y = 0;
    if (cursor_y < 0)             cursor_y = BOARD_SIZE_Y-1;

    toplevel = GetTopLayer(cursor_x,cursor_y);
    if (toplevel == -1) toplevel = 0;

    RestoreOldCursorPos();
    BackupCursorPos(toplevel,cursor_x,cursor_y);
    OutputCursor(toplevel,cursor_x,cursor_y);

    if (sel_layer == -1) pos = 0;
    else                 pos = 1;

    UpdateHelpField(GetStone(toplevel,cursor_x,cursor_y),pos);

    {
        char buf[100];
        sprintf(buf,"[F5 for Help]      Pairs Left = %d",pairs_left);

        DrawString(0,95,buf,F_4x6,A_REPLACE);
    }
}


unsigned char* helppage[] = {
    "To win this Game You have to remove all",
    "Stones from the Board.",
    "Stones can only be removed in Pairs and if the",
    "following Conditions hold true:",
    "",
    "(1) Both Stones have the same Graphics",
    "(2) Each Stone can be moved left OR right at",
    "    least one Field",
    "",
    "Use the Cursor Keys and [ENTER] or [F1] to",
    "select the Stones.",
    "",
    "[CLEAR] to Reset  [+][-] to change Setup",
    NULL};

/*===========================================================================*/
/* displays help page                                                        */
/*===========================================================================*/
void DisplayHelp() {
    int i = 0;

    running    = 0;
    outputtime = 0;

    memset(GetPlane(0),0,LCD_SIZE);
    memset(GetPlane(1),0,LCD_SIZE);


    DrawString(0,0,"TI-Mahjongg "MAHJONGG_VERSION" by TICT",F_6x8,A_REPLACE | A_SHADOWED);

    while (helppage[i]) {
        DrawString(0,10+i*7,helppage[i],F_4x6,A_REPLACE);
        i++;
    }
    GetUserInput();

    running    = 1;
    outputtime = 1;
}


/*===========================================================================*/
/* initialize a new game (returns 1 if okay)                                 */
/*===========================================================================*/
int InitGame(void) {
    int i=20;
    randomize();

    resettime = 1;
    running   = 1;
    cursor_x  = 0;
    cursor_y  = 0;

    //------------------------------------------------------------
    // for complicated setups we'll try at least 20 times to setup
    //------------------------------------------------------------
    do {
        if (!InitBoard()) return 0;
    } while (!CheckPossibleMovesLeft() && i--);

    if (i<=0) return 0;

    DrawBoard();
    PrintTime();
    CursorUpdate();
    return 1;
}


/*===========================================================================*/
/* draws a message                                                           */
/*===========================================================================*/
void DrawMessage(char* msg) {
    int sy = LCD_HEIGHT/2 - 2 + 44;

    DrawPopup(1,sy-6,LCD_WIDTH-1,sy+7);
    DrawString(0,sy-3-offsety,msg,F_6x8,A_REPLACE | A_CENTERED | A_SHADOWED);

    GetUserInput();
}


/*===========================================================================*/
/* handles and heavy error                                                   */
/*===========================================================================*/
void ErrorTreatment(void) {
    DrawMessage("ERROR: cannot initialize");
    FreeVariables();
    SetIntVec(AUTO_INT_5, oldint5);
    GrayMode(GRAY_OFF);
}


/*===========================================================================*/
/* as usual our main routine ...                                             */
/*===========================================================================*/
void _main(void) {
    int key;

    if (!GrayOn()) return;

    if (!AllocateVariables()) {
        GrayOff();
        FreeVariables();
        return;
    }

    outputtime = mseconds50 = seconds = minutes = hours = 0;

    Logo();

    if (!InitGame()) {
        DrawMessage("ERROR: cannot initialize");
        GrayOff();
        FreeVariables();
        return;
    }

    oldint5 = GetIntVec(AUTO_INT_5);
    SetIntVec(AUTO_INT_5, myint5handler);

    do {
        key = GetUserInput();

        if (key == KEY_UP) {
            cursor_y--;
            CursorUpdate();
        }
        else if (key == KEY_DOWN) {
            cursor_y++;
            CursorUpdate();
        }
        else if (key == KEY_LEFT) {
            cursor_x--;
            CursorUpdate();
        }
        else if (key == KEY_RIGHT) {
            cursor_x++;
            CursorUpdate();
        }
        else if (key == KEY_ENTER || key == KEY_F1) {
            if (HandleSelect(cursor_x,cursor_y)) {
                //---------------------------
                // !!! THE PLAYER HAS WON !!!
                //---------------------------
                char msg[100];
                running    = 0;
                outputtime = 0;
                sprintf(msg,"You've won in %02d:%02d:%02d !", hours, minutes, seconds);
                DrawMessage(msg);
                if (!InitGame()) {
                    ErrorTreatment();
                    return;
                }
            }
            else if (!CheckPossibleMovesLeft()) {
                //----------------------------
                // !!! THE PLAYER HAS LOST !!!
                //----------------------------
                char msg[100];
                running    = 0;
                outputtime = 0;
                sprintf(msg,"You lost (%d Pairs left)!",pairs_left);
                DrawMessage(msg);
                if (!InitGame()) {
                    ErrorTreatment();
                    return;
                }
            }
            else {
                CursorUpdate();
            }
        }
        else if (key == KEY_CLEAR || key == '+' || key == '-') {
            //------------------------
            // handle setup change ...
            //------------------------
            if (key == '+')      used_setup++;
            else if (key == '-') used_setup--;
            if (used_setup < 0)          used_setup = NR_SETUPS-1;
            if (used_setup >= NR_SETUPS) used_setup = 0;

            if (!InitGame()) {
                ErrorTreatment();
                return;
            }
        }
        else if (key == KEY_F5) {
            DisplayHelp();
            DrawBoard();
            CursorUpdate();
        }
    }
    while (key != KEY_ESC);

    SetIntVec(AUTO_INT_5, oldint5);
    FreeVariables();
    GrayOff();
}


//=============================================================================
// Revision History
//=============================================================================
//
// $Log: timah.c,v $
// Revision 1.1  2000/11/14 19:14:01  Thomas Nussbaumer
// initial version
//
//
